/*
 * Sonar, entreprise quality control tool.
 * Copyright (C) 2007-2008 Hortis-GRC SA
 * mailto:be_agile HAT hortis DOT ch
 *
 * Sonar is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * Sonar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with {library}; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02
 */
package ch.hortis.sonar.model;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.builder.EqualsBuilder;
import org.apache.commons.lang.builder.HashCodeBuilder;

import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.EnumType;
import javax.persistence.Enumerated;
import javax.persistence.FetchType;
import javax.persistence.GeneratedValue;
import javax.persistence.GenerationType;
import javax.persistence.Id;
import javax.persistence.JoinColumn;
import javax.persistence.ManyToOne;
import javax.persistence.OneToOne;
import javax.persistence.SequenceGenerator;
import javax.persistence.Table;

@Entity
@Table(name = "files")
public class File implements Collectable {
  
  public static final int NAME_COLUMN_SIZE = 256;
  public static final int NAMESPACE_COLUMN_SIZE = 500;
  public static final int PATH_COLUMN_SIZE = 500;

  @Id
  @Column(name = "id")
  @SequenceGenerator(name = "FILES_SEQ", sequenceName = "FILES_SEQ")
  @GeneratedValue(strategy = GenerationType.AUTO, generator = "FILES_SEQ")
  private Integer id;

  @Column(name = "filename", updatable = false, nullable = false, length = NAME_COLUMN_SIZE)
  private String filename;

  @Column(name = "namespace", updatable = false, nullable = true, length = NAMESPACE_COLUMN_SIZE)
  private String namespace;

  @Column(name = "path", updatable = false, nullable = true, length = PATH_COLUMN_SIZE)
  private String path;
  
  @Column(name = "file_type", updatable = false, nullable = false)
  @Enumerated(EnumType.STRING)
  private FileType type;

  @ManyToOne(fetch = FetchType.LAZY)
  @JoinColumn(name = "snapshot_id")
  private Snapshot snapshot;

  @OneToOne(mappedBy = "file", fetch = FetchType.LAZY, cascade = {CascadeType.ALL})
  public FileSource fileSource;

  public File(String filename, String namespace, String path) {
    this();
    setFilename(filename);
    setNamespace(namespace);
    setPath(path);
  }

  public File(int id, String filename, String namespace, String path) {
    this(filename, namespace, path);
    setId(id);
  }

  public File() {
    setType(FileType.MAIN);
  }

  public FileType getType() {
    return type;
  }

  public void setType(FileType type) {
    this.type = type;
  }

  public Snapshot getSnapshot() {
    return snapshot;
  }

  public void setSnapshot(Snapshot snapshot) {
    this.snapshot = snapshot;
  }

  public Integer getId() {
    return id;
  }

  public void setId(Integer id) {
    this.id = id;
  }

  public String getFilename() {
    return filename;
  }

  public void setFilename(String filename) {
    if ( filename.length() > NAME_COLUMN_SIZE ) {
      throw new IllegalArgumentException("File name " + filename + " size too long :" + filename.length());
    }
    String prefix = StringUtils.substringBeforeLast(filename, ".");
    String suffix = StringUtils.substringAfterLast(filename, ".");
    this.filename = prefix + "." + suffix.toLowerCase();
  }

  public String getPath() {
    return path;
  }

  public void setPath(String path) {
    if ( path != null && path.length() > PATH_COLUMN_SIZE ) {
      throw new IllegalArgumentException("File path " + path + " size too long :" + path.length());
    }
    this.path = path;
  }

  public String getNamespace() {
    return namespace;
  }

  public void setNamespace(String namespace) {
    if ( namespace != null && namespace.length() > NAMESPACE_COLUMN_SIZE ) {
      throw new IllegalArgumentException("File namespace " + namespace + " size too long :" + namespace.length());
    }
    this.namespace = namespace;
  }

  public FileSource getFileSource() {
    return fileSource;
  }

  public void setFileSource(FileSource fileSource) {
    this.fileSource = fileSource;
  }

  public boolean equals(Object obj) {
    if (!(obj instanceof File)) {
      return false;
    }
    if (this == obj) {
      return true;
    }
    File other = (File) obj;
    return new EqualsBuilder()
        .append(id, other.getId()).isEquals();
  }

  public int hashCode() {
    return new HashCodeBuilder(17, 37).
        append(id).toHashCode();
  }

  public String toString() {
    return id + ":".concat(namespace).concat(":").concat(filename);
  }

}
